###################################################################################################
#
#   (c) Copyright Motorola 2018. All rights reserved.
#   Motorola Confidential Proprietary
#   Contains confidential proprietary information of Motorola Mobility LLC.
#   Reverse engineering is prohibited.
#   The copyright notice does not imply publication.
#
###################################################################################################

import os

import ruamel.yaml

from davros_device_server.helper import user_path, which, windows
from davros_device_server.log_factory import setup_logger

logger = setup_logger("DavrosConfig")


class Config(object):
    CONFIG_FOLDER = "config"
    CONFIG_FILE = "davros.yaml"
    README = "README"
    VENV_FOLDER = os.path.join("venvs", "{version}")
    UPDATE_DAVROS_DEVICE_SERVER_FILE = "update_davros_device_server.py"

    @classmethod
    def davros_user_folder(cls):
        if windows():
            return os.path.join(user_path(), "AppData", "Local", "Davros")
        else:
            return os.path.join(user_path(), ".davros")
    @classmethod
    def davros_update_itself_path(cls):
            return os.path.join(cls.davros_user_folder(),cls.UPDATE_DAVROS_DEVICE_SERVER_FILE)
    @classmethod
    def venv_bin_path(cls, version):
        if windows():
            return os.path.join(
                cls.davros_user_folder(),
                cls.VENV_FOLDER.format(version=version),
                "Scripts",
            )
        else:
            return os.path.join(
                cls.davros_user_folder(), cls.VENV_FOLDER.format(version=version), "bin"
            )

    @classmethod
    def venv_activate_path(cls, version):
        return os.path.join(cls.venv_bin_path(version), "activate")

    @classmethod
    def venv_python_path(cls, version):
        return os.path.join(cls.venv_bin_path(version), "python")

    @classmethod
    def venv_pip_path(cls, version):
        return os.path.join(cls.venv_bin_path(version), "pip")

    @classmethod
    def venv_davros_path(cls, version):
        return os.path.join(cls.venv_bin_path(version), "davros-server")

    @classmethod
    def venv_path(cls, version):
        return os.path.join(
            cls.davros_user_folder(), cls.VENV_FOLDER.format(version=version)
        )

    @classmethod
    def venv_readme(cls):
        return os.path.join(cls.venvs(), cls.README)

    @classmethod
    def venvs(cls):
        return os.path.join(cls.davros_user_folder(), "venvs")

    @classmethod
    def config_path(cls):
        return os.path.join(cls.davros_user_folder(), "config")

    @classmethod
    def davros_pip_folder(cls):
        if windows():
            return os.path.join(user_path(), "AppData", "Roaming", "pip")
        else:
            return os.path.join(user_path(), ".config", "pip")

    @classmethod
    def davros_pip_file(cls):
        if windows():
            return "pip.ini"
        else:
            return "pip.conf"

    @classmethod
    def config_pip(cls, coreid, artifactory_key):
        if not os.path.isdir(cls.davros_pip_folder()):
            os.makedirs(cls.davros_pip_folder())
        with open(
                os.path.join(cls.davros_pip_folder(), cls.davros_pip_file()), "w"
        ) as f:
            f.write(
                "[global]\nfind-links = https://{}:{}@artifacts.mot.com/artifactory/api/pypi/pypi-local/simple\n"
                "index-url = https://{}:{}@artifacts.mot.com/artifactory/api/pypi/pypi-local/simple\n"
                "extra-index-url = https://pypi.org/simple/".format(
                    coreid, artifactory_key, coreid, artifactory_key
                )
            )
            f.close()

    @classmethod
    def load_config_file(cls):
        """
        Load davros.yaml config file

        :return: dict with the content of the config file read
        """
        try:
            yaml = ruamel.yaml.YAML(typ="safe")
            with open(os.path.join(cls.config_path(), cls.CONFIG_FILE), "r") as yaml_file:
                yaml_config = yaml.load(yaml_file)
            return yaml_config
        except Exception:
            logger.error("Error occurred while loading davros.yaml file")

    @classmethod
    def get_davros_prod(cls):
        """
        Get the master server URL from davros.yaml file
        """
        # production endpoint url is the default one
        master_server_endpoint_url = "https://davros.mot.com/kaleds/"
        try:
            davros_config = cls.load_config_file()
            master_server_endpoint_url = davros_config.get("MAIN", {}).get("DAVROS_PROD",
                                                                           "https://davros.mot.com/kaleds/")

            # adds a slash if it is not present
            if not master_server_endpoint_url.endswith('/'):
                master_server_endpoint_url += '/'

        except Exception:
            logger.error("Error occurred while loading DAVROS_PROD yaml property using the production one")

        return master_server_endpoint_url

    @classmethod
    def get_tardis_api_key(cls):
        """
        Get the tardis api key from davros.yaml file
        """
        # production endpoint url is the default one
        tardis_server_api_key = ""
        try:
            davros_config = cls.load_config_file()
            tardis_server_api_key = davros_config.get("MAIN", {}).get("DAVROS_API_KEY")
        except Exception:
            logger.error("Error occurred while loading DAVROS_API_KEY yaml property")

        return tardis_server_api_key

    @classmethod
    def get_tardis_ip(cls):
        """
        Get the tardis server IP address from davros.yaml file
        """

        tardis_server_ip = ""
        try:
            davros_config = cls.load_config_file()
            tardis_server_ip = davros_config.get("MAIN", {}).get("SERVER_IP_ADDRESS")
        except Exception:
            logger.error("Error occurred while loading SERVER_IP_ADDRESS yaml property")

        return tardis_server_ip

    @classmethod
    def get_environment_type(cls):
        """
        Get the environment type from davros.yaml.example file
        """
        # production is the default environment
        environment_type_value = "PROD"
        try:
            davros_config = cls.load_config_file()
            environment_type_value = davros_config.get("MAIN", {}).get("ENVIRONMENT_TYPE", "PROD")
        except Exception:
            logger.error("Error occurred while loading ENVIRONMENT_TYPE yaml property using the production one")

        return environment_type_value

    @classmethod
    def create_davros_conf(cls, fields):
        """ Creates the TARDIS config file if it doesn't exist."""
        if not os.path.isdir(Config.config_path()):
            os.makedirs(Config.config_path())

        Config.config_pip(fields[0], fields[1])
        davros_config = {
            "MAIN": {
                "ARTIFACTORY_API_KEY": fields[1],
                "ARTIFACTORY_USERNAME": fields[0],
                "DAVROS_PROD": "https://davros.mot.com/kaleds/",
                "FRAME_TIMEOUT": "3",
                "IS_TARDIS": "false",
                "LOG_LEVEL": "DEBUG",
                "LOG_STD_OUT": "true",
                "MAX_FRAMERATE_PER_SECOND": "20",
                "SERVER_OWNER": fields[0],
            }
        }
        if fields[2]:
            davros_config["MAIN"]["PLATFORM_TOOLS_PATH"] = fields[2]

        yaml = ruamel.yaml.YAML()
        with open(os.path.join(Config.config_path(), "davros.yaml"), "w") as fp:
            yaml.dump(davros_config, fp)
