###################################################################################################
#
#   (c) Copyright Motorola 2018. All rights reserved.
#   Motorola Confidential Proprietary
#   Contains confidential proprietary information of Motorola Mobility LLC.
#   Reverse engineering is prohibited.
#   The copyright notice does not imply publication.
#
###################################################################################################

import json
import os
import signal
import socket
import stat
import subprocess
from datetime import timedelta
from distutils.version import Version
from tempfile import gettempdir
from time import time

import pkg_resources
import requests

import davros_device_server.davros_server_manager
from davros_device_server.davros_config import Config
from davros_device_server.davros_server_manager import DavrosServerManager
from davros_device_server.job import Job
from davros_device_server.log_factory import setup_logger


class AutoUpdater:
    """
    Base class for auto updater jobs.

    :param str custom_path:

    """

    PID_TIMEOUT = 30
    CHECK_TIME = 120

    def __init__(self, wizard=False, custom_path=None):
        self.pid = os.getpid()
        self._pid_file = os.path.join(gettempdir(), "davros_device_server.pid")
        if not custom_path:
            self._davros_path = Config.davros_user_folder()
        else:
            self._davros_path = custom_path
        self.davros_pid = None

        self._logger = setup_logger(self.__class__.__name__)
        self.server_manager = DavrosServerManager(custom_path, wizard)
        self.create_job()
        self._is_running = False

        def version_callback(version, status):
            self._logger.info("Version: {}, Status: {}".format(version, status))

        self.version_callback = None if wizard else version_callback

    def create_job(self):
        """ Creates the updater job for upgrading TARDIS version."""
        self.update_job = Job(timedelta(seconds=self.CHECK_TIME), self.update_davros)

    def update_davros(self):
        """ Update TARDIS version based on master server endpoint."""
        self._update_running_file()
        self.update_itself()
        if not self.server_manager.update_server():
            # Warns about the error and try to update again next time.
            self._logger.warning(f"Failed to update tardis server. " +
                                 "This may be related to problems to reach artifactory and/or Davros.")

            davros_device_server.davros_server_manager.retries += 1
            if davros_device_server.davros_server_manager.retries <= 5:
                self._logger.warning("Retrying connection... number of retries: "
                                     f"{davros_device_server.davros_server_manager.retries}")
            else:
                self._logger.info("Number of retries exceeded!")
                raise SystemError("Error in server's environment...")
        else:
            davros_device_server.davros_server_manager.retries = 0

    def get_version_update_info(self):
        """ Retrieves the latest Tardis version from Davros and checks if it can be updated."""
        davros_version = DavrosVersions.get_latest_version()

        return davros_version, True

    def start_autoupdater(self, force):
        """
        Start autoupdater instance or restart if it is running.
        """

        if force:
            self._logger.info("Another Autoupdater instance was detected. Killing it...")
            self.server_manager.kill_davros(force)

        # send auto-updater-version to master server
        self._send_updater_info_to_master()
        return self.start_loop(force)

    def stop_autoupdater(self, force):
        """ Stop auto-updater execution by killing its process."""
        self.server_manager.kill_davros(force)
        self.stop_loop()

    def _update_running_file(self):
        """ Update/Create the PID file containing the PID number."""
        with open(self._pid_file, "w") as fh:
            fh.write(str(self.pid))
        self._set_file_permissions()
        os.utime(self._pid_file, None)
        self._is_running = True

    def _set_file_permissions(self):
        """ Set PID file permissions."""
        permission = (
            stat.S_IROTH
            | stat.S_IWOTH
            | stat.S_IRUSR
            | stat.S_IWUSR
            | stat.S_IRGRP
            | stat.S_IWGRP
        )
        os.chmod(self._pid_file, permission)

    def _kill_running(self):
        """ Kills auto-updater the running process number based onthe PID file."""
        try:
            with open(self._pid_file, "r") as pid_f:
                pid = int(pid_f.read())
            os.kill(pid, signal.SIGTERM)
            self._logger.info("Killing Autoupdater instance...")
        except Exception as e:
            self._logger.info("Autoupdater is not running.")
            self._logger.info(e)

    def start_server(self, force):
        """
        Start a Tardis server instance.
        """
        if not self.server_manager.check_dependencies():
            self._logger.info("Dependencies missing. Server will not start until fixed.")
            return

        if self.server_manager.davros_system_version() is None:
            self._logger.info(
                "There is no Tardis instance installed. Installing the latest version..."
            )
            self.server_manager.update_server()
        else:
            if not force and self.server_manager.get_davros_processes():
                self._logger.info(
                    "There is another Tardis instance running. If you want to restart Tardis anyway, "
                    "add the 'force' option"
                )
            else:
                self._logger.info("Starting Tardis server.")
                self.server_manager.restart_davros()

    def _running_pid(self):
        """
        Check if an Autoupdater instance is running on the current machine.
        """
        try:
            last_update = int(time() - os.path.getmtime(self._pid_file))
            with open(self._pid_file, "r") as pid_f:
                pid = int(pid_f.read())
            if self.PID_TIMEOUT * 2 > last_update:
                return pid
            else:
                return -1
        except (FileNotFoundError, ValueError):
            return -1

    def _check_running(self, kill=False):
        """
        Checks if autoupdater is running.

        :param bool kill: If True and the autoupdater is running, it tries to kill it.
        :return: An autoupdater instance is running
        """
        pid = self._running_pid()
        if int(pid) > 0:
            if kill:
                self._logger.info("Killing another autoupdater instance.")
                try:
                    os.kill(pid, signal.SIGINT)
                except ProcessLookupError as e:
                    self._logger.error("Process already finished.")
                    pass
                return False
            else:
                return True
        else:
            self._logger.info("Another Autoupdater instance detected.")
            return False

    def start_loop(self, force=False):
        """
        Starts an Autoupdater looping routine.

        :param bool force: If another is instance is running, kills it a start a new one.
        :return: True if the loop was successfully started.
        """
        running = self._check_running(kill=force is True)
        if running:
            return False

        self._update_running_file()
        self.update_job.start()
        self._setup_signals()
        return True

    def stop_loop(self, signum=None, frame=None):
        """ Stop checking if there is a new TARDIS version available."""
        self._logger.info(
            "Autoupdater {} has been stopped with signal {}.".format(frame, signum)
        )

        self.update_job.stop()
        if not self._is_running:
            self._kill_running()

    def _setup_signals(self):
        """ Setup killing signals to stop auto-updater."""
        signal.signal(signal.SIGTERM, self.stop_loop)
        signal.signal(signal.SIGINT, self.stop_loop)

    def first_installation(self):
        """ Return if the config path exists."""
        return not os.path.exists(Config.config_path())

    def _send_updater_info_to_master(self):
        """
        Sends autoupdater info to master.
        """
        hostname = socket.gethostname()

        # Checks if there is the parameter SERVER_IP_ADDRESS in Yaml file, if it isn't gets the ip via socket
        ip = Config.get_tardis_ip()
        if not ip:
            ip = socket.gethostbyname(hostname)

        # join the master server url with the auto-updater version endpoint
        url = Config.get_davros_prod() + "davros/api/tardis/auto-updater/version"

        # get the auto-updater version
        version: Version = pkg_resources.require("davros_device_server")[0].version
        version_str = version.__str__()

        # build the JSON to be sent
        json = {
            "hostname": hostname,
            "ip": ip,
            "version": version_str
        }

        # set request header
        headers = {
            "X-Davros-Api-Key": Config.get_tardis_api_key()
        }
        req = requests.put(url=url, json=json, headers=headers, verify=False if Config.get_environment_type() == "DEV"
                            else True, timeout=5)
        self._logger.info(f"VERSION PERSISTENCE: Request finished with status code {req.status_code}")

    def update_itself(self):
        auto_updater_version = self.check_itself_version()
        if not auto_updater_version is None:
            f = os.popen("pip show davros-device-server")
            details = f.read().splitlines()
            f.close()
            for d in details:
                if 'Version' in d:
                    current_version = d.split(':')[1].strip()
                    break
            if current_version < auto_updater_version:
                self._logger.info("Start to update itself to {} version".format(auto_updater_version))
                try:
                    self.run_update_itself_script(auto_updater_version)
                except Exception as e:
                    self._logger.error(str(e))
            else:
                self._logger.info("The auto-updater version: {} is the latest one,no need to update".format(current_version))
        else:
            self._logger.error("The auto-updater version is None,please double check")
    def run_update_itself_script(self,version):
         davros_update_itself_path = Config.davros_update_itself_path()
         subprocess.Popen(
                [
                    davros_update_itself_path,
                    "-v {}".format(version)
                ],
                shell=True
            ).wait()
    def check_itself_version(self):
        """
        Checks the auto-updater server version that the Master requires the Slave Server to be

        :return str: auto-updater server version Master requires
        """

        _MASTER_ENDPOINT = Config.get_davros_prod() + "davros/api/auto-update/checkSelf"
        _HEADERS = {"Content-type": "application/json"}
        _PARAMS = {"deviceServerHostname": socket.gethostname() or socket.getfqdn()}

        current_version = None
        try:
            response = requests.get(
                _MASTER_ENDPOINT,
                headers=_HEADERS,
                params=_PARAMS,
                verify=False if Config.get_environment_type() == "DEV" else True,
                timeout=10,
            )
            data = json.loads(response.content)
            if data and len(data.get("current_version").strip()) != 0:
                current_version = data.get("current_version")
        except Exception as e:
            self._logger.error(str(e))
        return current_version
