import json
import os
import re
import shutil
import socket
import subprocess
from os import listdir, remove
from os.path import exists, getmtime, isdir, join
from tempfile import gettempdir
from time import ctime, sleep
from urllib.parse import urljoin

import psutil
import requests
from packaging.version import InvalidVersion, Version

from davros_device_server.davros_config import Config
from davros_device_server.helper import windows
from davros_device_server.log_factory import setup_logger

READMETEXT = """
This folder contains instalations of Davros server, DO NOT REMOVE OR TAMPER WITH THEM.
"""

retries = 0


class ADBError(Exception):
    """
    Exception raised for unexpected ADB errors

    :param error: original error from ADB
    :type error: string or exception
    """

    def __init__(self, error):
        super(ADBError, self).__init__("ADB error: '{}'".format(str(error)))


class FastbootError(Exception):
    """
    Exception raised for unexpected fastboot errors

    :param error: original error from fastboot
    :type error: string or exception
    """

    def __init__(self, error):
        super(FastbootError, self).__init__("Fastboot error: '{}'".format(str(error)))


def _check_adb():
    """
    Checks if the system has ADB installed.
    """
    try:
        out = subprocess.check_output("adb devices", shell=True)
    except ADBError:
        return False
    return True


def _check_fastboot():
    """
    Checks if the system has Fastboot installed.
    """
    try:
        out = subprocess.check_output("fastboot devices", shell=True)
    except FastbootError:
        return False
    return True


class VersionMatchException(Exception):
    MSG = "Virtual environment version ({venv_version}) is different than the package version installed"

    def __init__(self, venv_version):
        super(VersionMatchException, self).__init__(
            self.MSG.format(venv_version=venv_version)
        )


class DavrosServerManager(object):

    def __init__(self, custom_path=None, wizard=False):
        self._logger = setup_logger("DavrosServerManager")
        self.pypi_version = None
        self.running_version = None
        if not custom_path:
            self._davros_path = Config.davros_user_folder()
        else:
            self._davros_path = custom_path

        self._wizard = wizard
        self.davros_pid = None
        self.davros_process = None

    def get_davros_processes(self):
        """
        Check if a davros server instance is running on the current machine.
        """
        procs = []
        for proc in psutil.process_iter():
            try:
                if "davros_server" in proc.cmdline() or "davros-server" in proc.name():
                    procs.append(proc)
            except (PermissionError, psutil.AccessDenied, psutil.NoSuchProcess):
                pass
        return procs

    def restart_davros(self, version=None):
        """
        Start davros server instance or restart if one is running.
        :return bool: True if tardis is successfully restarted; False, otherwise.
        """
        davros_version = version or self.davros_system_version()

        self._logger.info("Starting Tardis server.")

        if davros_version is None:
            self._logger.error("No Tardis version is installed yet.")
            return False

        self.kill_davros(force=True)

        self._start_server(davros_version)
        return True

    def kill_davros(self, force=True):
        """
        Kill davros server instance.
        """
        procs = self.get_davros_processes()
        for p in procs:
            try:
                self._logger.info(f"Davros server instance detected: PID: {p.pid} - killing it.")
                children = p.children()
                children.append(p)

                for c in children:
                    try:
                        c.terminate()
                    except psutil.NoSuchProcess as e:
                        self._logger.info(f"Process {e.pid} already killed. Skipping...")
                        pass
                gone, alive = psutil.wait_procs(children, timeout=300,
                                                callback=lambda process: print("Process killed:", process))
                for a in alive:
                    a.kill()
            except psutil.NoSuchProcess as e:
                self._logger.info(f"Davros process of PID {e.pid} no longer exists. Skipping...")
                pass

        self.davros_process = None

    def start_server(self):
        """
        Start an davros server instance.
        """
        davros_version = self.davros_system_version()
        self._logger.info("Starting davros server")

        if davros_version is None:
            self._logger.error("No version of davros is installed yet.")
            return

        self._start_server(davros_version)

    def check_dependencies(self):
        """ Checks if the ADB and Fastboot are available."""
        if not _check_adb():
            self._logger.error("ADB instance was not found.")
            return False
        if not _check_fastboot():
            self._logger.error("Fastboot instance was not found.")
            return False
        return True

    def get_installed_versions(self):
        """
        Returns all versions installed with modification dates, does a simple check if it is a virtual env also
        Obs.: Files that are not a virtualvenv or the README will be deleted.

        :return: Installed version with time of modification.
        :rtype: List of tuple
        """

        def is_venv(path):
            venv_bin = "Scripts" if windows() else "bin"
            if exists(join(path, venv_bin, "activate")):
                return True
            return False

        path = Config.venvs()

        all_files = [join(path, p) for p in listdir(path)]

        if Config.venv_readme() in all_files:
            all_files.remove(Config.venv_readme())
        else:
            with open(Config.venv_readme(), "w") as f:
                f.write(READMETEXT)

        self._logger.debug(f"Venv candidates found: {all_files}")
        venvs = []
        for version_candidate in all_files:
            if isdir(version_candidate):
                if is_venv(version_candidate):
                    venvs.append((version_candidate, getmtime(version_candidate)))
                else:
                    self._logger.info(
                        f"Removing folder {version_candidate} from venv folders."
                    )
                    try:
                        shutil.rmtree(version_candidate)
                    except Exception as e:
                        self._logger.error(
                            f"Couldn't remove folder {version_candidate}: {e}"
                        )
            else:
                self._logger.info(
                    f"Removing file {version_candidate} from venv folders."
                )
                try:
                    remove(version_candidate)
                except Exception as e:
                    self._logger.error(f"Couldn't remove file {version_candidate}: {e}")

        self._logger.debug(f"Found versions: {[(v, ctime(t)) for v, t in venvs]}")
        return venvs

    def clean_old_versions(self, remain=5):
        """
        Cleans the vevnv folder of old versions of davros installed, leaving only newest versions

        :param int remain: number of version that will remain.
        """
        versions_installed = self.get_installed_versions()
        versions_installed = sorted(versions_installed, key=lambda v: v[1])

        remove = versions_installed[: -1 * remain]
        self._logger.debug(f"Deleting old versions: {remove}")
        for venv, _ in remove:
            try:
                shutil.rmtree(venv)
            except Exception as e:
                self._logger.error(f"Couldn't remove venv {venv}: {e}")

    def _start_server(self, davros_version):
        """
        Start an davros server instance.
        """

        self.clean_old_versions()
        python_path = Config.venv_python_path(davros_version)
        output = open((os.path.join(gettempdir(), "output.txt")), "w")
        self.kill_davros(force=True)
        self.davros_process = subprocess.Popen(
            [python_path, "-m", "davros_server", "start"],
            cwd=self._davros_path,
            stdout=output,
            stderr=output,
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP if windows() else 0,
        )

        if self.davros_process.poll() is None:
            self._logger.info(
                "Davros server {} has been started with pid {}.".format(
                    davros_version, self.davros_process.pid
                )
            )
            self.version_callback(davros_version, "running")
            self.running_version = davros_version
        else:
            self._logger.error(
                "An error occurred when initializing Davros. Davros server could not be started."
            )

    def version_callback(self, version, status):
        if self._wizard:
            self._logger.info("Version: {}, Status: {}".format(version, status))

    def create_venv(self, version):
        """
        Create a clean virtualenv for install davros package.
        :param str version: Version being installed
        """
        self._logger.info("Creating venv for davros version {}.".format(version))
        if not os.path.exists(Config.venvs()):
            self._logger.info(
                "First time using davros: Creating virtualenv path... {}".format(
                    self._davros_path
                )
            )
            os.makedirs(Config.venvs())
        try:
            new_venv_path = Config.venv_path(version)
            p = subprocess.Popen(
                [
                    "python.exe"
                    if windows()
                    else "python",
                    "-m",
                    "virtualenv",
                    new_venv_path,
                ],
                stdout=subprocess.PIPE,
            )
            p.wait()
        except Exception as e:
            self._logger.error(
                "Error creating virtualenv for version {} {}".format(version, e)
            )

    def _install_server(self, version):
        """
        Install a brand new version of davros-server package hosted on motorola's pypi repository into
        a previous created python virtual environment.
        :param str version: Version to be installed
        :return bool success: If installation process was successful
        """
        pip_path = Config.venv_pip_path(version)
        try:
            self._logger.info(
                "Updating pip in server virtualenv for version '{}'".format(version)
            )
            subprocess.Popen(
                [pip_path, "--isolated", "--timeout=20", "install", "-U", "pip"]
            ).wait()
            self._logger.info(
                "Installing davros server in virtualenv for version '{}'".format(
                    version
                )
            )
            subprocess.Popen(
                [
                    pip_path,
                    "install",
                    "--timeout=180",
                    "davros_server=={}".format(version),
                ]
            ).wait()
        except Exception as e:
            self._logger.error("A problem happened while installing the server")
            self._logger.exception(e)
            return False

        if self.check_davros_installation(version):
            self._logger.info("Davros {} was successfully installed.".format(version))
            return True
        else:
            self._logger.error(
                "Problem installing version {} of davros server.".format(version)
            )
            return False

    def update_server(self):
        """
        Try to update davros-server package to latest available version on motorola's pypi repository.
        :return bool: True if it was possible to update tardis/it's already running the latest version; False, otherwise.
        """
        self._logger.info("Checking for updates...")

        # Step 1: Check what is the latest version stored on Davros endpoint
        pypi_version, can_update = self.check_master_required_version()
        print("pypi_version:----------------------")
        print(pypi_version)
        print("can_update:----------------------")
        print(can_update)
        if not pypi_version:
            self._logger.error(
                "It was not possible to fetch latest Tardis server version from Davros server!"
            )
            return False
        else:
            self.pypi_version = pypi_version

        # Step 2: Check if given version is already installed. If not, install it by creating a venv
        if not self.check_davros_installation(pypi_version):
            self._logger.info(
                "Installing version {} required by Master".format(pypi_version)
            )
            self.create_venv(pypi_version)
            global retries
            retries = 0
            if not self._install_server(self.pypi_version):
                try:
                    shutil.rmtree(Config.venv_path(self.pypi_version))
                except Exception as e:
                    self._logger.error(
                        "Error during installation of latest version of Tardis server: {}".format(e)
                    )
                    self._logger.error(
                        "Please check your artifactory credentials on pip.conf."
                    )
                    raise SystemError("Error in server's environment...")
                return False
            # Wait some time to davros settle PID file
            sleep(5)

        # Step 3: Check if davros-server has a running process. If not, (re)start it.
        print("pypi_version:----------------------")
        print(pypi_version)
        print("self.running_version:----------------------")
        print(self.running_version)
        if self.get_davros_processes():
            print("pypi_version:----------------------")
            print(pypi_version)
            print("self.running_version:----------------------")
            print(self.running_version)
            if can_update and pypi_version != self.running_version:
                
                self._logger.info(f"Restarting Tardis server with new version {pypi_version}")
                return self.restart_davros(pypi_version)
        else:
            self._logger.info("Tardis server isn't running yet! Starting...")
            return self.restart_davros(pypi_version)
        return True

    def davros_system_version(self):
        """
        Check if has some version installed on the system and which is it.
        """
        try:
            sys_version = self.get_latest_version()
        except VersionMatchException:
            sys_version = None

        return sys_version

    def check_davros_installation(self, version):
        """
        Check if davros is installed inside a specific virtualenv version.
        :param path: Path to desire virtualenv.
        :param version: Virtualenv version.
        :return: Davros installation version.
        """
        try:
            pip_freeze = subprocess.Popen(
                [Config.venv_pip_path(version), "freeze"], stdout=subprocess.PIPE
            )
            p_out, err = pip_freeze.communicate()
        except OSError:
            self._logger.error("Could not use virtualenv to check Davros installation.")
            return None
        if not err:
            decode_output = p_out.decode()
            find_package = decode_output.find("davros-server")
            if find_package != -1:
                parser = decode_output[find_package:]
                parser = parser.split("\n")[0]
                return parser.split("==")[1]
        return None

    def get_latest_version(self):
        """
        Get latest version installed on current machine.
        :param path: Path to desire Davros installation.
        """
        venv_versions = []
        # check if davros user folder exists, if not, create user a venv folders
        if not os.path.isdir(Config.venvs()):
            os.makedirs(Config.venvs())
        for v in os.listdir(Config.venvs()):
            try:
                venv_versions.append(Version(v))
            except InvalidVersion:
                self._logger.error("{} is not a valid folder version.".format(v))
        if not venv_versions:
            return None
        latest_version = max(venv_versions)
        # check if virtualenv version matches with davros package version
        if not self.check_davros_installation(latest_version):
            raise VersionMatchException(latest_version)
        return latest_version

    def check_newer_version(self):
        """
        Check the latest davros slave server version on motorola pypi server.
        It parses the latest server version using installing error message from 'pip install davros-server=='.
        :return Version.version: Latest davros-server version
        """

        try:
            pip_freeze = subprocess.Popen(
                ["pip", "install", "--timeout=5", "davros-server=="],
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            _, err = pip_freeze.communicate()
            decode_err = err.decode()
            package_versions = [x for x in re.split(r"[()]", decode_err) if x.strip()]
            package_versions = package_versions[1].replace("from versions: ", "")
            package_versions = [
                package_versions.strip()
                for package_versions in package_versions.split(",")
            ]

            version = Version(package_versions[-1])

        except Exception as e:
            self._logger.error(
                "Fail to parse latest davros slave server version: {}.".format(e)
            )
            return None
        return version

    def check_master_required_version(self):
        """
        Checks the version that the Master requires the Slave Server to be

        :return str: server version Master requires
        :return bool: Can device_server update slave at the moment
        """

        _MASTER_ENDPOINT = Config.get_davros_prod() + "davros/api/auto-update/check"
        _HEADERS = {"Content-type": "application/json"}
        _PARAMS = {"deviceServerHostname": socket.gethostname() or socket.getfqdn()}

        current_version = None
        can_update = False

        try:
            response = requests.get(
                _MASTER_ENDPOINT,
                headers=_HEADERS,
                params=_PARAMS,
                verify=False if Config.get_environment_type() == "DEV" else True,
                timeout=10,
            )
            data = json.loads(response.content)

            if data:
                current_version = data.get("current_version")
                if current_version:
                    current_version = Version(current_version)
                can_update = data.get("can_update", False)
        except Exception as e:
            self._logger.error(str(e))

        return current_version, can_update
